@@ -48,6 +48,8 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 #include <NetworkManager.h>
 
+#include <pulse/pulseaudio.h>
+
 #define PAGE_INTRO 0
 #define PAGE_LOCALE 1
 #define PAGE_PASSWD 2
@@ -408,7 +410,7 @@ static gboolean show_ip (void);
 static gboolean net_available (void);
 static int get_pi_keyboard (void);
 static gboolean srprompt (gpointer data);
-static void uscan_toggle (GtkSwitch *sw, gpointer ptr);
+static void uscan_toggle (GtkSwitch *sw, gpointer, gpointer);
 
 /* Helpers */
 
@@ -2356,7 +2358,7 @@ static gboolean close_prog (GtkWidget *widget, GdkEvent *event, gpointer data)
 
 /* Underscan */
 
-static void uscan_toggle (GtkSwitch *sw, gpointer ptr)
+static void uscan_toggle (GtkSwitch *sw, gpointer, gpointer)
 {
     int enable = gtk_switch_get_active (sw) ? 0 : 1;
     if (GTK_WIDGET (sw) == uscan2_sw)
@@ -2373,6 +2375,128 @@ static int num_screens (void)
         return get_status ("xrandr -q | grep -cw connected");
 }
 
+/* PulseAudio control */
+
+typedef struct {
+    pa_threaded_mainloop *pa_mainloop;
+    pa_context_state_t pa_state;
+    char *pa_default_sink;
+} pulse_data_t;
+
+static void pa_cb_state (pa_context *pacontext, void *userdata)
+{
+    pulse_data_t *pdata = (pulse_data_t *) userdata;
+
+    if (pacontext == NULL) pdata->pa_state = PA_CONTEXT_FAILED;
+    else pdata->pa_state = pa_context_get_state (pacontext);
+
+    pa_threaded_mainloop_signal (pdata->pa_mainloop, 0);
+}
+
+static void pa_cb_generic_success (pa_context *context, int success, void *userdata)
+{
+    pulse_data_t *pdata = (pulse_data_t *) userdata;
+
+    pa_threaded_mainloop_signal (pdata->pa_mainloop, 0);
+}
+
+static void pa_cb_get_sink_info (pa_context *context, const pa_sink_info *info, int eol, void *userdata)
+{
+    pulse_data_t *pdata = (pulse_data_t *) userdata;
+
+    if (!eol)
+    {
+        if (info->description && !strncmp (info->description, "Built-in Audio", 14))
+        {
+            // if no device set, take the first one found...
+            if (!pdata->pa_default_sink) pdata->pa_default_sink = g_strdup (info->name);
+            else if (strstr (pdata->pa_default_sink, "stereo-fallback"))
+            {
+                // ... if the set device is analogue and another has been found, it is HDMI - use it instead
+                g_free (pdata->pa_default_sink);
+                pdata->pa_default_sink = g_strdup (info->name);
+            }
+        }
+    }
+
+    pa_threaded_mainloop_signal (pdata->pa_mainloop, 0);
+}
+
+#define PA_OP_WAIT \
+    if (op) \
+    { \
+        while (pa_operation_get_state (op) == PA_OPERATION_RUNNING) pa_threaded_mainloop_wait (pdata.pa_mainloop); \
+        pa_operation_unref (op); \
+    }
+
+static void set_audio_output (void)
+{
+    pa_context *pa_cont;
+    pa_proplist *paprop;
+    pa_mainloop_api *paapi;
+    pa_operation *op;
+    pa_cvolume pacvol;
+    pulse_data_t pdata;
+    int i;
+
+    pdata.pa_mainloop = pa_threaded_mainloop_new ();
+    if (pdata.pa_mainloop == NULL) goto pa_terminate;
+
+    pa_threaded_mainloop_start (pdata.pa_mainloop);
+    pa_threaded_mainloop_lock (pdata.pa_mainloop);
+
+    paprop = pa_proplist_new ();
+    pa_proplist_sets (paprop, PA_PROP_APPLICATION_NAME, "unknown");
+    pa_proplist_sets (paprop, PA_PROP_MEDIA_ROLE, "music");
+    paapi = pa_threaded_mainloop_get_api (pdata.pa_mainloop);
+    pa_cont = pa_context_new_with_proplist (paapi, "unknown", paprop);
+    pa_proplist_free (paprop);
+    if (pa_cont == NULL) goto pa_terminate;
+
+    pdata.pa_state = PA_CONTEXT_UNCONNECTED;
+    pa_context_set_state_callback (pa_cont, &pa_cb_state, &pdata);
+    pa_context_connect (pa_cont, NULL, PA_CONTEXT_NOAUTOSPAWN, NULL);
+    while (pdata.pa_state != PA_CONTEXT_READY && pdata.pa_state != PA_CONTEXT_FAILED)
+    {
+        pa_threaded_mainloop_wait (pdata.pa_mainloop);
+    }
+    if (pdata.pa_state != PA_CONTEXT_READY) goto pa_terminate;
+
+    // find sinks
+    pdata.pa_default_sink = NULL;
+    op = pa_context_get_sink_info_list (pa_cont, &pa_cb_get_sink_info, &pdata);
+    PA_OP_WAIT
+    if (pdata.pa_default_sink == NULL) goto pa_terminate;
+
+    // set the default sink
+    op = pa_context_set_default_sink (pa_cont, pdata.pa_default_sink, &pa_cb_generic_success, &pdata);
+    PA_OP_WAIT
+
+    // unmute the sink
+    op = pa_context_set_sink_mute_by_name (pa_cont, pdata.pa_default_sink, 0, &pa_cb_generic_success, &pdata);
+    PA_OP_WAIT
+
+    // set the default sink volume
+    pacvol.channels = PA_CHANNELS_MAX;
+    for (i = 0; i < pacvol.channels; i++) pacvol.values[i] = 65535;
+    op = pa_context_set_sink_volume_by_name (pa_cont, pdata.pa_default_sink, &pacvol, &pa_cb_generic_success, &pdata);
+    PA_OP_WAIT
+
+pa_terminate:
+    if (pdata.pa_mainloop != NULL)
+    {
+        if (pa_cont != NULL)
+        {
+            pa_context_disconnect (pa_cont);
+            pa_context_unref (pa_cont);
+        }
+
+        pa_threaded_mainloop_unlock (pdata.pa_mainloop);
+        pa_threaded_mainloop_stop (pdata.pa_mainloop);
+        pa_threaded_mainloop_free (pdata.pa_mainloop);
+    }
+    if (pdata.pa_default_sink) g_free (pdata.pa_default_sink);
+}
 
 /* The dialog... */
 
@@ -2399,7 +2523,7 @@ int main (int argc, char *argv[])
     if (!init_user) init_user = g_strdup ("pi");
 
     // set the audio output to HDMI if there is one, otherwise the analog jack
-    vsystem ("hdmi-audio-select");
+    set_audio_output ();
 
     // read country code from Pi keyboard, if any
     kbd = get_pi_keyboard ();
@@ -2483,8 +2607,8 @@ int main (int argc, char *argv[])
     uscan2_box  = (GtkWidget *) gtk_builder_get_object (builder, "p7hbox2");
     gtk_switch_set_active (GTK_SWITCH (uscan1_sw), !get_status ("raspi-config nonint get_overscan_kms 1"));
     gtk_switch_set_active (GTK_SWITCH (uscan2_sw), !get_status ("raspi-config nonint get_overscan_kms 2"));
-    g_signal_connect (uscan1_sw, "state-set", G_CALLBACK (uscan_toggle), NULL);
-    g_signal_connect (uscan2_sw, "state-set", G_CALLBACK (uscan_toggle), NULL);
+    g_signal_connect (uscan1_sw, "notify::active", G_CALLBACK (uscan_toggle), NULL);
+    g_signal_connect (uscan2_sw, "notify::active", G_CALLBACK (uscan_toggle), NULL);
     if (num_screens () != 2) gtk_widget_hide (uscan2_box);
 
     // set up the locale combo boxes
